/*!
 *  \file main.c
 *  \brief oWL Pico reference design software.
 *
 *  Copyright (C) 2010 HD Wireless AB
 *
 *  You should have received a copy of the license along with this library.
 */


/*! \defgroup wlp_ref oWL Pico Reference Design
 *
 * The pclient application is built on top of the oWL pico API and uses the
 * owl toolkit, both provided by H&D Wireless.
 *
 * pclient provides a command line interface to directly invoke most of the
 * API functions in oWL Pico API, a ttcp throughput test tool and a small
 * web server.
 *
 * pclient can be built and used on a linux desktop PC together with the
 * SPB800 device. It can also be built for an ATMEL Xplain board, running on an
 * 8-bit CPU with 8Kb RAM. All the platform differences are handled in the
 * owl driver layer and in the main.c file.
 *
 * To build pclient on a linux system, make -C ports/linux should produce the
 * pclient binary in the ports/linux dir.
 *
 * To build pclient on for the ATMEL Xplain board, make -C ports/avr8 should
 * produce the pclient.hex binary in the ports/avr8 dir.
 *
 * The pclient reference design can be used as a starting point for using the
 * oWL Pico API for a custom application and/or on a custom platform.
 * 
 *  @{
 */
#define OWL_TRACE OWL_TRACE_INFO

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>

#ifdef __linux__
# include <sys/time.h>
# include <fcntl.h>
# include <unistd.h>
#endif /* __linux__ */

#ifdef __avr8__
# include <avr/io.h>
# define F_CPU 32000000UL
# include <util/delay.h>
# define UART_SHELL   USARTC0
# define UART_PICO    USARTD0
#endif /* __avr8__ */

#include <wlp_api.h>
#include <wlp_inet.h>

#include <owl/core/owl_err.h>
#include <owl/core/owl_debug.h>
#include <owl/core/owl_shell.h>
#include <owl/core/owl_timer.h>
#include <owl/drivers/owl_board.h>
#include <owl/drivers/owl_uart.h>
#include <owl/drivers/owl_rtc.h>
#include <owl/drivers/owl_led.h>
#include <owl/drivers/owl_temp.h>
#include "wlp_shell.h"
#include "wlp_ttcp.h"
#include "wlp_ntp.h"
#include "wlp_ping.h"
#include "httpd/wlp_httpd.h"

#define TIMER_HZ 10

struct httpd *http_handle = NULL;

#ifdef __linux__
static int kbhit(void)
{
        struct timeval tv;
        fd_set fds;
        tv.tv_sec = 0;
        tv.tv_usec = 0;
        FD_ZERO(&fds);
        FD_SET(0, &fds);
        select(0 + 1, &fds, NULL, NULL, &tv);
        return FD_ISSET(0, &fds);
}


static void stdout_outb(void *ctx, int c)
{
        putc(c, stdout);
        fflush(stdout);
}


void owl_putc(char c)
{
        putc(c, stderr);
}
#endif /* __linux__ */


#ifdef __avr8__
void owl_putc(char c)
{
        owluart_outb(&UART_SHELL, c);
}
#endif /* __avr8__ */


static void link_cb(void *ctx, int link)
{
        struct owlsh_ctx *sh = ctx;
        owlsh_printf(sh, "link_cb link:%d\n\r", link);
}


static void addr_cb(void *ctx, const struct ip_addr *ip)
{
        struct owlsh_ctx *sh = ctx;
        if (ip == NULL) {
                owlsh_printf(sh, "addr_cb ip:none\n\r");
                if (http_handle) {
                        httpd_destroy(http_handle);
                        http_handle = NULL;
                        owlsh_printf(sh, "httpd stopped\n\r");
                }

        } else {
                owlsh_printf(sh, "addr_cb ip:%s\n\r", inet_ntoa(ip));
                if ((http_handle = httpd_create(80)) != NULL)
                        owlsh_printf(sh, "httpd started\n\r");
        }
}


static void tick_cb(void)
{
	owltmr_tick();
}


int main(int argc, char *argv[])
{
        int err;
        struct owlsh_ctx *sh;
        int demo_mode = 0; /* we use an optional demo mode on xplain (avr8) */
        
        owlboard_init();
        owlrtc_init(TIMER_HZ, tick_cb);
        owltemp_init();

        if ((err = owltmr_init(TIMER_HZ)) < 0) {
                owl_err("owltmr_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }
        
        /* We will use the owl_* portability layer for console
         * input/output and error/debug functionality. This allows us
         * to build this code on (for instance) a Linux PC system
         * during development. To build the code for the target 
         * host platform you need to 
         *  - Write a owl_env.h file for the host
         *  - Write a owl_uart.c file for the target platform, implementing
         *    the functions prototyped in owl_uart.h
         *  - Write a owl_led.c file for the target platform, implementing
         *    the functions prototyped in owl_led.h
         *  - Write a owl_rtc.c file for the target platform, implementing
         *    the functions prototyped in owl_rtc.h
         */

#ifdef __avr8__

        /* On the xplain board, skip demo mode if user pressed any button
         * during startup
         */
        demo_mode = 1;
        if (PORTF_IN != 0xff) /* XXX should live in owl_button.c */
                demo_mode = 0;

        /* Initialize the console UART */
        if (owluart_init(&UART_SHELL, UART_BAUDRATE_9600,
                         UART_DATABITS_8, UART_PARITY_NONE,
                         UART_STOPBITS_1, 0) < 0)
                return -1;
        
        /* Initialize the host-WiFi UART */
        if ((err = owluart_init(&UART_PICO, UART_BAUDRATE_57600,
                                UART_DATABITS_8, UART_PARITY_NONE,
                                UART_STOPBITS_1, 0)) < 0) {
                owl_err("uart_init pico err:%d(%s)", err, owl_error(err));
                return -1;
        }

        /* Initialize the host console input/output system. We will
         * only use the console for status output in the example
         * (using owlsh_print()) so the console command handling will
         * be unused.
         */
        if ((sh = owlsh_create(NULL, owluart_outb, &UART_SHELL,
                               OWLSH_ESCAPECHARS)) == NULL) {
                owl_err("owlsh_create failed");
                return -1;
        }

        /* Wait for wifi to start */
        _delay_ms(200);

        /* Setup the owl pico library. We pass the functions
         * that the library will use to read from and write to
         * the WiFi module.
         */
        if ((err = wlp_init(115200, 0,
                            owluart_read, owluart_write, &UART_PICO)) < 0) {
                owl_err("wlp_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }
        
        /* Initialize the host-WiFi UART */
        if ((err = owluart_init(&UART_PICO, UART_BAUDRATE_115200,
                                UART_DATABITS_8, UART_PARITY_NONE,
                                UART_STOPBITS_1, 0)) < 0) {
                owl_err("uart_init pico err:%d(%s)", err, owl_error(err));
                return -1;
        }

        _delay_ms(100);
        
#endif /* __avr8__ */
        
#ifdef __linux__
        int fd;
        char *devname;
        int baudrate, rtscts;
        
        if (argc > 1 && !strcmp(argv[1], "-h")) {
                printf("%s [port] [baudrate] [rtscts]\n", argv[0]);
                return 0;
        }

        if(argc > 1)
                devname = argv[1];
        else
                devname = "/dev/ttyUSB0";

        if (argc > 2)
                baudrate = atol(argv[2]);
        else
                baudrate = 115200;

        if (argc > 3)
                rtscts = atol(argv[3]);
        else
                rtscts = 0;

        if ((fd = open(devname, O_RDWR | O_NOCTTY)) < 0) {
                owl_err("Failed to open device '%s'", devname);
                return -1;
        }
        
        if ((err = owluart_init(&fd, UART_BAUDRATE_57600,
                                UART_DATABITS_8, UART_PARITY_NONE,
                                UART_STOPBITS_1, 0)) < 0) {
                owl_err("uart_init failed err:%d(%s)", fd, owl_error(fd));
                return -1;
        }

        if ((sh = owlsh_create(NULL, stdout_outb, NULL,
                               OWLSH_NOECHO | OWLSH_ESCAPECHARS)) == NULL) {
                owl_err("owlsh_create failed");
                return -1;
        }

        if ((err = wlp_init(baudrate, rtscts,
                            owluart_read, owluart_write, (void *) fd)) < 0) {
                owl_err("wlp_init failed err:%d(%s)", err, owl_error(err));
                return -1;
        }

        sleep(1);

        owl_info("opening %s baudrate:%d rtscts:%d", devname, baudrate, rtscts);
        if ((err = owluart_init(&fd, baudrate,
                                UART_DATABITS_8, UART_PARITY_NONE,
                                UART_STOPBITS_1, rtscts)) < 0) {
                owl_err("uart_init failed err:%d(%s)", fd, owl_error(fd));
                return -1;
        }

        owlsh_addcmd(sh, &cmd_send_file, NULL);
        owlsh_addcmd(sh, &cmd_recv_file, NULL);
        owlsh_addcmd(sh, &cmd_recv, NULL);
#endif /* __linux__ */

        /* We register callbacks so that we get notified when the
         * WiFi link is established (link up) or torn down (link down)
         * and when and IP address has been assigned to the interface
         * (since this can happen at some point after the WiFi link
         * comes up if DHCP is used).
         */
        wlp_set_link_cb(link_cb, sh);
        wlp_set_ipaddr_cb(addr_cb, sh);
        
        /* Add commands which can be used to control the WiFi link.
         * link_cb() will be called when the connection is up.
         */
        owlsh_addcmd(sh, &cmd_linkup, NULL);
        owlsh_addcmd(sh, &cmd_linkdown, NULL);
        owlsh_addcmd(sh, &cmd_get_hwaddr, NULL);
        owlsh_addcmd(sh, &cmd_get_network, NULL);

        /* Add commands which can be used to control the ip address settings */
        owlsh_addcmd(sh, &cmd_set_ipaddr, NULL);
        owlsh_addcmd(sh, &cmd_get_ipaddr, NULL);
        owlsh_addcmd(sh, &cmd_set_dhcp, NULL);
        owlsh_addcmd(sh, &cmd_get_dhcp, NULL);
        owlsh_addcmd(sh, &cmd_get_hostbyname, NULL);

        /* Add commands which can be used to play around with the socket
         * interface of owl pico
         */
        owlsh_addcmd(sh, &cmd_socket, NULL);
        owlsh_addcmd(sh, &cmd_bind, NULL);
        owlsh_addcmd(sh, &cmd_listen, NULL);
        owlsh_addcmd(sh, &cmd_connect, NULL);
        owlsh_addcmd(sh, &cmd_close, NULL);
        owlsh_addcmd(sh, &cmd_send, NULL);
        owlsh_addcmd(sh, &cmd_sendto, NULL);
        owlsh_addcmd(sh, &cmd_get_peeraddr, NULL);
        
#ifdef __linux__
        /* These commands require a proper timer implemenation.
         * Since timers are not yet implemented on avr8, we only support these
         * commands on a linux host in this release.
         */
        owlsh_addcmd(sh, &cmd_ttcp, NULL);
        owlsh_addcmd(sh, &cmd_ntp, NULL);
        owlsh_addcmd(sh, &cmd_ping, NULL);
#endif
        owlsh_addcmd(sh, &cmd_help, NULL);

        /* We use a default configuration on xplain for demo purposes */
        if (demo_mode) {
                struct ip_addr ip, netmask;
                struct wl_ssid_t ssid; /* default ssid */

                IP4_ADDR(&ip, 192, 168, 2, 22);
                IP4_ADDR(&netmask, 255, 255, 255, 0);

                strcpy(ssid.ssid, "hdwireless");
                ssid.len = strlen("hdwireless");

                /* connect to default AP */
                wlp_linkup(&ssid, "hdwireless123", 0);

                /* set default IP address */
                wlp_set_ipaddr(&ip, &netmask, NULL, NULL);
        }
        
        /* Start the main poll loop to check the console input and making sure
         * that the owl pico API makes forward progress.
         */
        for (;;) {

#ifdef __avr8__
                if (owluart_poll(&UART_SHELL))
                        owlsh_key(sh, owluart_inb(&UART_SHELL, 0));
#endif /* __avr8__ */

#ifdef __linux__
                if (kbhit()) {
                        char line[80];
                        char *p;

                        fgets(line, sizeof(line), stdin);
                        p = line;
                        while (*p) {
                                owlsh_key(sh, *p);
                                p++;
                        }
                }
#endif /* __linux__ */
                
                wlp_poll();
                owltmr_poll();
        }                
        
        return 0;
}

/*!  @} */
